                               (* BLOCKMO.PAS *)

unit BlockMo;

(* Gibbs sampling estimation for multi-block model;
   programmed by Tom A.B. Snijders, University of Groningen,
   transformed to Windows by Peter Boer, ProGAMMA.
   Refers to paper "Estimation and prediction for stochastic blockstructures"
   by Krzysztof Nowicki and Tom Snijders.
   (JASA, september 2001.)
   Most symbols in the program are identical to those in the paper.
   Alpha = set of types of relations
   AA = script A = alphabet of pairwise relations.
   The unit BLOCKMO contains the procedures ReadInfile and Readdata
   for reading the basic information file and the data file,
   together with some related variables and general procedures. *)


interface

Uses
   WBlib, RanGen, B_cons, BlockI, SysUtils;
(* Blib is a general procedure library;
   RanGen is a library of random variable generators;
   B_cons contains constants;
   BlockI contains procedures for user interaction. *)


type

Blockmodel = object

proname:         string;                               (* Project name  *)
a:               array[1..rmax,1..2] of shortint;      (* Alphabet *)
c,r0,r1:         integer;                              (* c = number of colors *)
n,ndat:          integer;                              (* n= number of vertices;
                                                          ndat = Number of dyads with observed values *)
y:               array of array of shortint;   (* Observations *)
data:            array of array of Boolean;    (* true for non-missing data *)
ppi:             array[1..rmax] of integer;            (* Reflection operator pi *)
tekst:           string;                               (* name text for identification *)
clusternumber:   integer;                               (* for Pajek output *)
oldcode1, oldcode2: array[0..cmax] of shortint; (* original data codes for the new color alphabet *)


comin, comax, missing:            integer;
(* minimum and maximum numbers of colors; missing data indicator *)
nrun0,nrun1:                      longint;
(* number of runs before and after convergence is assumed *)
ranseed:                          longint;
(* Random number seed *)
theta00:    real;
(* theta and eta are parameters;
   theta0 and eta0 are their prior values.
   theta0 = T_k in the paper,
   and defined as theta0 = 1 + theta00 * c (where c = number of colors).
   Note that theta0 is identical over all coordinates. *)
calcsd, identified, both, IdentificationByFile:
                       boolean;
(* calcsd indicates whether standard errors are to be computed;
   identified indicates whether the model is identified
   by prior probabilities for x;
   both indicates that first a non-identified model, and then
   an identified model is to be run;
   IdentificationByFile indicates that prior probabilities
   were read from a file. *)
multiple, optimize, randoptimize, overdisx, overdiseta :
                        boolean;
(* multiple indicates more than one Gibbs sequence per model;
   optimize indicates optimization of starting values;
   randoptimize indicates optimization of starting values as in versions before 1.6;
   overdisx and overdiseta indicate overdispersion of x and of eta
   in the first half of the pre-convergence iterations. *)
multi:         integer;
(* multi = total number of Gibbs sequences per model
           if option multiple is chosen *)
colver:                array[1..cmax] of integer;
lastthreshold:         real;
(*  in the "both" option,
  colver[i] is the vertex that can be used for identifying color i, and
  lastthreshold indicates how good the separation for the last color is *)
prior:                 array of Boolean;
(* prior[i] indicates whether there is a prior probability for vertex i *)
thetai:                array of array of real;
(* if prior[i], then theta[i,j] gives the prior probabilities of x[i]=j *)
v:              array[1..cmax] of integer;
(* identifying vertices in the "identified, mode = 1" option *)

procedure SetArrays(var outf: textfile);

function r01: integer;
function r:   integer;
function symbol(i: integer): string;

procedure Heading(var outf: textfile; level: integer; s: string);
procedure initialise(var outf, outf1: textfile; var outfn, outfn1: string);
procedure ReadInfile(var outf, outf1: textfile; var outfn, outfn1: string);
procedure readdata(var outf, outf1: textfile; outfn, outfn1: string);
procedure IdentifyVertices(var outf: textfile);

end; (* object Blockmodel *)

var version:        string;
  dfn, pfn:       string; (* data file name; prior probabilities file name *)

implementation

procedure Blockmodel.SetArrays(var outf: textfile);
var i,j: integer;
begin
  SetLength(y,n+1,n+1);
  SetLength(data,n+1,n+1);
  Setlength(prior,n+1);
  SetLength(thetai,n+1,cmax+1);
  for i:=1 to n do begin
    prior[i] := false;
    for j:=1 to cmax do thetai[i,j] := 0.0;
   end;
end; (* setarrays *)


function Blockmodel.r01: integer;
begin
  r01 := r0+r1;
end;

function Blockmodel.r: integer;
begin
  r := r0+2*r1;
end;

function Blockmodel.symbol(i: integer): string;
function st(i: integer):string;
(* The string containing the number i . *)
var ss: string;
begin
  str(i:3,ss);
  st := ss;
end; (* st *)
begin
  symbol := '('+st(a[i,1])+','+st(a[i,2])+')';
end;

procedure Blockmodel.Heading(var outf: textfile; level: integer; s: string);
var i: integer;
begin
  writeln(outf);
  writeln(outf,'@',level);
  writeln(outf,s);
  if (length (s) > 0)
    then for i:=length(s) downto 1 do
    case level of
      1: write(outf,'*');
      2: write(outf,'*');
      3: write(outf,'=');
      4: write(outf,'=');
      5: write(outf,'-');
      6: write(outf,'.');
      7: write(outf,'.');
     end; (* case *)
  writeln(outf);
  writeln(outf);
end;


procedure Blockmodel.initialise(var outf, outf1: textfile;
                                var outfn, outfn1: string);
var  sd: longint;
begin
  optMessage3('Initialise project');
  attachproject(outf,outf1,proname,outfn,outfn1);
  writeln(outf);
  writeln(outf,'BLOCKS version ',version);
  writeln(outf);
  clusternumber := 0;
end; (* initialise *)

procedure Blockmodel.ReadInfile(var outf, outf1: textfile;
                                var outfn, outfn1: string);
var inf, df, pf: textfile;

 infn: string;
 mode, dumm, i, j: integer;
 finished, impossible: boolean;
 sum: real;

procedure parsename(var f: textfile; var fn: string);
var started, terminated: boolean;
 ch: char;
begin (* parsename *)
(* Reads a string enclosed by blanks or EndOfLine,
   and omits any characters with ASCII numbers not between 43 and 126. *)
  fn := '';
  started := false;
  terminated := false;
  repeat
   read(f,ch);
   if (((ch = ' ') or Eoln(f)) and started) then terminated := true;
   if (ch <> ' ') then begin
     started := true;
     if ((ord(ch) >= 43) and (ord(ch) <= 126)) then fn := fn+ch;
    end;
   until (terminated or Eoln(f));
  readln(f);
 end; (* procedure parsename *)

begin (* procedure ReadInfile *)
(* read basic information file *)
  infn := proname+'.in';
  if not FileExists(infn) then begin
    writeln(outf,'The present directory does not contain a file named ',infn,',');
    writeln(outf,'which is the expected name for the basic information file.');
    stopthis('Information file not found.',outf, outf1, outfn, outfn1);
    exit;
   end;
  assign(inf,infn);
  reset(inf);
  OptMessage3('Read project description file');
  Heading(outf,1,'Input of project description file');
  writeln(outf,'Read project description from basic information file <',
              infn,'>.');
  writeln(outf);

  readln(inf,tekst);
  writeln(outf,'Title line of basic information file is ');
  writeln(outf,tekst,'.');
  DefineTitle(tekst);
  parsename(inf,dfn);
  writeln(outf,'Data file is named <',dfn,'>.');
  if (dfn = '') then begin
    writeln(outf,'The first line of file ',infn,' does not contain an allowable name');
    writeln(outf,'for the data file.');
    stopthis('Data file name not found on first line of basic information file.',
             outf, outf1, outfn, outfn1);
    exit;
   end;
  if not FileExists(dfn) then begin
    writeln(outf,'The present directory does not contain this data file.');
    stopthis('Data file not found.',outf, outf1, outfn, outfn1);
    exit;
   end;
(* Determine number of vertices *)
  writeln(outf,'Preliminary inspection of first line of the data file:');
  assign(df,dfn);
  reset(df);
  n := 0;
  repeat
    inc(n);
    if (n > 999999) then begin
       writeln(outf,'More than ',999999,' numbers on line 1 of the data file.');
       writeln(outf,'This seems so strange that the program stops.');
       stopthis('Too many vertices in this data file.',
                outf, outf1, outfn, outfn1);
       exit;
      end; (* if n > 999999 *)
    read(df,dumm);
   until eoln(df);
  if (n <= 2) then begin
       writeln(outf,'Only ',n,' numbers on line 1 of the data file.');
       writeln(outf,'This leads to nothing.');
       stopthis('Not enough vertices in this data file.',
                outf, outf1, outfn, outfn1);
       exit;
   end;
  writeln(outf,'Data file contains ',n,' vertices.');
  closefile(df);
  writeln(outf);

  writeln(outf,'The following options were chosen: ');
  readln(inf,comin,comax);
  if (comin <= 1) then begin
    writeln(outf,'Minimum number of colors given as ',comin,'.');
    writeln(outf,'This makes no sense. Changed to 2.');
    comin := 2;
   end;
  if (comax < comin) then begin
    writeln(outf,'Maximum number of colors given as ',comax,'.');
    writeln(outf,'It must be at least as large as the minimum number.');
    writeln(outf,'Changed to 2.');
    comax := comin;
   end;
  if (comax > cmax) then begin
    writeln(outf,'Maximum number of colors given as ',comax,'.');
    writeln(outf,'It must be less than constant cmax, which is ',cmax,'.');
    writeln(outf,'Changed to ',cmax,'.');
    comax := cmax;
   end;

  write(outf,'Number of colors to be analyzed');
  if (comin = comax) then writeln(outf,' is ',comin,'.')
   else writeln(outf,' ranges from ',comin,' to ',comax,'.');

  readln(inf,nrun0);
  if (nrun0 < 100) then begin
    writeln(outf,'Number of pre-convergence runs given as ',nrun0,'.');
    writeln(outf,'This is too small. Changed to 100.');
    nrun0 := 100;
   end;
  writeln(outf,'Default number of pre-convergence runs is ',nrun0,'.');

  readln(inf,nrun1);
  if (nrun1 < 100) then begin
    writeln(outf,'Number of post-convergence runs given as ',nrun1,'.');
    writeln(outf,'This is too small. Changed to 100.');
    nrun1 := 100;
   end;
  writeln(outf,'Default number of post-convergence runs is ',nrun1,'.');

  readln(inf,multi);     
  writeln(outf,'Code for multiple Gibbs sequences is ',multi,'.');
  if (multi < 0) then begin
    writeln(outf,'This is not allowed. Changed into 0.');
    multi := 0;
   end;
  if (multi > mulmax) then
    begin
    writeln(outf,'The maximum number in this implementation is ',mulmax,'.');
    multi := mulmax;
    end;
  multiple := (multi >= 2);
  if (multi = 0) then begin
    writeln(outf,'No Gibbs sequences requested, only data inspection.');
   end else begin
    write(outf,multi,' Gibbs sequence');
    if multiple then write(outf,'s');
    writeln(outf,' for each color.');
   end;

  readln(inf,mode);
   case mode of
     0: begin identified := false; both := false;
        writeln(outf,'Non-identified model.'); end;
     1: begin identified := true;  both := false;
        writeln(outf,'Identified model.');  end;
     2: begin identified := true;  both := false;
        writeln(outf,'Identified model.'); end;
     3: begin identified := false; both := true;
        writeln(outf,'Non-documented option (code 3):');
        writeln(outf,'First run a non-identified model,');
        writeln(outf,
  'then a model identified from the results of the non-identified model.');
         end;
      else begin
         writeln(outf,'Code for identification mode given as ',mode,'.');
         writeln(outf,
         'This is not permitted. Changed into 0 (non-identified model).');
         mode := 0;
         identified := false; both := false;
       end;
    end; (* case mode *)
  IdentificationByFile := (mode = 2);

  readln(inf,missing);
  writeln(outf,'Missing value indicator ',missing,'.');

  readln(inf,theta00);
  if (theta00 < 0.0) then begin
    writeln(outf,'Concentration parameter given as ',theta00:20:10,' .');
    writeln(outf,'This parameter must be nonnegative. Changed into 0.');
    theta00 := 0.0;
   end;
  writeln(outf,'Concentration parameter ',theta00:8:2,' .');

  readln(inf,dumm);
  optimize := (dumm >= 1);
  randoptimize := (dumm = 9);

  readln(inf,dumm);
  overdisx := (dumm >= 1);

  readln(inf,dumm);
  overdiseta := (dumm >= 1);

  if optimize then writeln(outf,'Search for good starting values.')
   else writeln(outf,'Random starting values.');
  if randoptimize then begin
    writeln(outf,'Non-documented option (code 9):');
    writeln(outf,'good starting values are randomly sought (like in versions before 1.6).');
   end;
  if (overdisx or overdiseta) then begin
    write(outf,'First half of the pre-convergence iterations ');
    write(outf,'with overdispersed');
   end;
  if overdisx then begin
    write(outf,' colorings');
    if overdiseta then write(outf,' and') else write(outf,'.');
   end;
  if overdiseta then writeln(outf,' eta probabilities.');
  writeln(outf); writeln(outf);

  for i:=1 to cmax do v[i] := 0;
  case mode of
    0: (* nothing to do *) ;
    1: begin
           i := 0;
           repeat begin
             i := i+1;
             read(inf,v[i]);
            end
           until ((i >= comax) or (i >= cmax) or (v[i] = 0) );
           writeln(outf,'Vertices for identifying the colors : ');
           i := 0;
           repeat
             i := i+1;
             if (v[i] > 0) then begin
               if (i > 1) then writeln(outf,', ');
               write(outf,'color ',i,' identified by vertex ',v[i]);
              end;
            until ((i >= comax) or (i >= cmax) or (v[i] = 0));
           writeln(outf,'.');
         end; (* case 1 *)
     2: begin
        parsename(inf,pfn);
        writeln(outf,'Prior probabilities file name <',pfn,'>.');
        if (pfn = '') then begin
          writeln(outf,'Line 6 of file ',infn,' indicates that mode = ',mode,',');
          writeln(outf,'but the 12th line of file ',infn,' does not contain an allowable name');
          writeln(outf,'for the file with prior probabilities.');
          stopthis('No name for fil with prior probabilities.',
                   outf, outf1, outfn, outfn1);
          exit;
         end;
        if not FileExists(pfn) then begin
          writeln(outf,'The present directory does not contain a data file named ',pfn,'.');
          stopthis('Data file with identifying probabilities not found.',outf, outf1, outfn, outfn1);
          exit;
         end;
        writeln(outf);
        writeln(outf,'Read prior probabilities from data file ',pfn,'.');
        writeln(outf,'Contents of this file : ');
        assign(pf,pfn);
        reset(pf);
        i := 0;
        finished := false;
        repeat
          i := i+1;
          read(pf,dumm);
          if eof(pf) then finished := true else begin
            if ((dumm <= 0) or (dumm > n)) then begin
              writeln(outf,'Line ',i,' of file ',pfn,' starts with the number ',dumm,'.');
              writeln(outf,'This is not allowed for a vertex number.');
              stopthis('Incorrect number in prior probabilities file.',
                       outf, outf1, outfn, outfn1);
              exit;
             end;
            sum := 0.0;
            impossible := false;
            for j:=1 to comin do begin
              read(pf,thetai[dumm,j]);
              if (thetai[dumm,j] < -1e-32) then impossible := true;
              sum := sum + thetai[dumm,j];
              if (eof(pf) and (j < comin)) then finished := true;
             end;
            if (abs(sum - 1.0) > 1e-10) then impossible := true;
            if (not finished) then begin
              prior[dumm] := true;
              readln(pf);
              write(outf,dumm:3,' ');
              for j :=1 to comin do write(outf,thetai[dumm,j]:6:4,' ');
              writeln(outf);
              if (impossible) then begin
                writeln(outf,'Line ',i,' of file ',pfn,
                 ' does not contain a correct vector of probabilities.');
   stopthis('Incorrect probability vector in file with prior probabilities.',
                  outf, outf1, outfn, outfn1);
                exit;
               end;
             end; (* if not finished *)
           end; (* if not eof *)
         until ((i > n - 2) or eof(pf) or finished);
        closefile(pf);
       end; (* case 2 *)
    3: (* nothing to do *) ;
   end; (* case mode *)

(* Now follows in the .in file the 13th line indicating whether the output
   file is to be appended or rewritten. This was read already in
   procedure attachproject in unit BlockI. *)

  readln(inf);
  if eof(inf) then
    begin
    writeln(outf,'error in basic information line: no line with random number seed.');
    writeln(outf,'The default value 0 is used.');
    ranseed := 0;
    end
   else
    begin
    readln(inf,ranseed);
    writeln(outf,'Random number seed read as ',ranseed,'.');
    end;
  if (ranseed = 0) then
    writeln(outf,'This means that the program is randomized by the clock.');
(* Initialise random number generator *)
  if (ranseed = 0) then
    begin
    randomize;
    ranseed := trunc(random*1234567);
    writeln(outf,'Generated random number generator seed is ',ranseed,'.');
    end;
  init3(ranseed);
  writeln(outf);


  closefile(inf);

  if ((multi = 0) and (mode = 1)) then
   for i:= comin to comax do begin
    writeln(outf);
    write(outf,'For ',i, ' colors :');
    c := i;
    IdentifyVertices(outf);
   end;

  if (comax > cmax) then begin
    writeln(outf,'The maximum number of colors is ',comax,'.');
    writeln(outf,'This is more than constant cmax = ', cmax,' in B_CONS.');
    writeln(outf,'Recompile BLOCKS with larger constant cmax.');
    stopthis('Too many relation types in this data.',
             outf, outf1, outfn, outfn1);
    exit;
   end; (* if *)
end; (* ReadInfile *)

procedure Blockmodel.readdata(var outf, outf1: textfile; outfn, outfn1: string);
var
  df: textfile;
  present: array[-cc..cc,-cc..cc] of integer;
  number: array[1..rmax] of longword;
  i,j,nmis: integer;
  ya,yb: shortint;
  founderror : Boolean;
procedure tellalphabet(var f: textfile);
var i: integer;
  ent,fract: real;
begin
  if (r0 > 0) then begin
    writeln(f,'Set of symmetric tie values is ');
    for i:=1 to r0 do writeln(f,i:2,' ',symbol(i),'   (',number[i]:4,' dyads)');
   end else writeln(f,'No symmetric tie values.');
  if (r1 > 0) then begin
    writeln(f,'Set of asymmetric tie values is ');
    for i:=1 to 2*r1 do begin
      write(f,r0+i:2,' ',symbol(r0+i));
      writeln(f,'   (',number[r0+i]+number[ppi[r0+i]]:4,' dyads)');
      if (i = r1) then
        if (r1 = 1) then writeln(f,'and its mirrored counterpart')
                    else writeln(f,'and their mirrored counterparts');
     end;
   end else writeln(f,'No asymmetric tie values.');
  writeln(f,'Total number of tie values is ',r,'.'); writeln(f);
(*  writeln(f,'Reflection function pi is');
  writeln(f,'  a   pi(a) ');
  for i:=1 to r do writeln(f,i:3,ppi[i]:6); writeln(f); *)
  ndat := (sqr(n) - nmis) div 2;
  writeln(f,'Number of dyads with observed values is ',ndat,'.');
  writeln(f,'Number of dyads with missing  values is ',(nmis -n) div 2,'.');
  if (nmis - n >=2) then
    writeln(f,'Code for missing data in new alphabet is 0.');

  ent := 0.0;
  for i:=1 to r0 do begin
    fract := number[i]/ndat;
    ent := ent - fract*log(fract);
   end;
  for i:=1 to r1 do begin
    fract := (number[r0+i] + number[ppi[r0+i]])/ndat;
    ent := ent - fract*log(fract);
   end;
  writeln(f);
  writeln(f,'Information calculated with natural logarithms.');
  writeln(f,'Entropy of empirical dyad distribution = ',ent:10:5);
  writeln(f);
end; (* tell alphabet *)

begin (* readdata *)
(* Read data file *)
  Heading(outf,1,'Input of data file');
  Heading(outf1,1,'Input of data file');
  OptMessage3('Read data file');
  assign(df,dfn);
  reset(df);
  writeln(outf,'Data read from file ', dfn,' and reproduced on file ',outfn1,'.');
  writeln(outf1,'Input data matrix from file ',dfn,' is '); writeln(outf1);
(* First determine the number of numbers on the first data line *)
  n := 0;
  repeat
    inc(n);
    read(df,y[1,n]);
    write(outf1,' ',y[1,n]:2);
   until eoln(df);
  writeln(outf,n,' vertices');
  writeln(outf1);
(* Read remainder of data matrix *)
  for i:=2 to n do begin
    for j:=1 to n do begin
      read(df,y[i,j]);
      if (eof(df) and (not ((i=n) and (j=n)))) then begin
        writeln(outf);
        writeln(outf,
        'Unexpected end of file when trying to read data element y[',i,',',j,'].');
        writeln(outf,'There is something wrong with the data file. Check the format.');
        stopthis('Unexpected end of data file.',outf, outf1, outfn, outfn1);
        exit;
       end;
      write(outf1,' ',y[i,j]:2);
     end; (* j *)
    writeln(outf1);
   end; (* i *)
  writeln(outf1); writeln(outf);
  closefile(df);

(* Make the alphabet *)
(* Present first indicates whether this combination of digits
   is present in the data set *)
  OptMessage3('Make alphabet');
  nmis := 0;
  founderror := false;
  for i:=-cc to cc do for j:=-cc to cc do present[i,j] := 0;
  for i:=1 to rmax do ppi[i] := 0;
  for i:=1 to n do for j:=1 to n do begin
    if not ((y[i,j] >= -cc) and (y[i,j] <=cc)) then begin
      writeln(outf,'Value ',j,' in data row ',i,' of file ',dfn,' is ',y[i,j],'.');
      writeln(outf,'However, all values should be between ',-cc,' and ',cc,'.');
      founderror := true;
     end;
    if ((i <> j) and (y[i,j] <> missing) and (y[j,i] <> missing))
      then present[y[i,j],y[j,i]] := 1 else inc(nmis);
   end; (* for i for j *)
  if founderror then begin
    writeln(outf,'Change your data file or your program.');
    stopthis('Value out of range in data file.',outf, outf1, outfn, outfn1);
    exit;
   end;
(* Now present will be given new values,
   indicating the new codes for the values in the data set.
   The new codes are the numbers in the alphabet A *)
  r0 := 0;
  for i := -cc to cc do if (present[i,i] >= 1) then begin
    inc(r0);
    ppi[r0] := r0;
    if (r0 > rmax) then begin
      writeln(outf,'Number of possible symmetric tie values is larger than ',rmax,'.');
      writeln(outf,'This requires a recompilation of the program.');
      stopthis('Too many symmetric tie values in this data file.',
               outf, outf1, outfn, outfn1);
      exit;
     end;
    a[r0,1] := i;
    a[r0,2] := i;
    present[i,i] := r0;
   end; (* if *)
  r1 := 0;
  for i := -cc to cc do for j:=i+1 to cc do if (present[i,j] >= 1) then begin
    inc(r1);
    if (r > rmax) then begin
      writeln(outf,'Total number of possible dyadic tie values is larger than ',rmax,'.');
      writeln(outf,'This requires a recompilation of the program.');
      stopthis('Too many relations in this data file.',
               outf, outf1, outfn, outfn1);
      exit;
     end;
    a[r0+r1,1] := i;
    a[r0+r1,2] := j;
    present[i,j] := r0+r1;
   end;
  for i:=1 to r1 do begin
    ppi[r0+i] := r0+r1+i;
    ppi[r0+r1+i] := r0+i;
    for j:=1 to 2 do a[r0+r1+i,j] := a[r0+i,3-j];
    present[a[r0+i,2],a[r0+i,1]] := r0+r1+i;
   end;
  if (r01 > r01max) then begin
    writeln(outf,'Total number r01 of non-redundant possible dyadic tie values is larger than ',r01max,'.');
    writeln(outf,'This requires a recompilation of the program.');
    stopthis('Too many non-redundant relations in this data file.',
             outf, outf1, outfn, outfn1);
    exit;
   end;

(* Retain old codes for later use: invert array present: *)
  for i:=0 to cmax do
    begin
    oldcode1[i] := 0;
    oldcode2[i] := 0;
    end;
  for i:=-cc to cc do for j:=-cc to cc do
    begin
    if (not (present[i,j] in [0..cmax])) then
      begin
      writeln(outf);
      write(outf,'Error: present[',i,',',j,'] = ', present[i,j],',');
      writeln(outf,' but this value should be between 0 and cmax = ',cmax);
      stopthis('Range check error.',outf, outf1, outfn, outfn1);
      exit;
      end;

    oldcode1[present[i,j]] := i;
    oldcode2[present[i,j]] := j;
    end;

(* Recode data matrix to the new alphabet *)
  OptMessage3('Recode data matrix');
  for i:=1 to n do begin for j:=1 to i-1 do begin
    if ((y[i,j] <> missing) and (y[j,i] <> missing))
      then begin
        ya := y[i,j]; yb := y[j,i];
        y[i,j] := present[ya,yb];
        y[j,i] := present[yb,ya];
        data[i,j] := true;
        data[j,i] := true;
       end (* if *) else (* missing *) begin
        y[i,j] := 0;
        y[j,i] := 0;
        data[i,j] := false;
        data[j,i] := false;
       end;
     end; (* j *)
    y[i,i] := 0;
    data[i,i] := false;
   end; (* i *)

  writeln(outf1); writeln(outf1,'Recoded data matrix is '); writeln(outf1);
  for i:=1 to n do begin
    for j:=1 to n do begin
      write(outf1,' ',y[i,j]:2);
     end; (* j *)
    writeln(outf1);
   end; (* i *)
  writeln(outf1);

(* Finally, numbers is used to count the number of the dyads with the
   various relation types *)
   
  OptMessage3('Count dyads');
  for i:=1 to rmax do number[i] := 0;
  for i:=1 to n do for j:=1 to i-1 do
    if data[i,j] then inc(number[y[i,j]]);

  tellalphabet(outf);

 if (r <= 1) then begin
   writeln(outf,'The number of relation types is ',r,'.');
   writeln(outf,'For less than 2 relations, estimation makes no sense.');
   writeln(outf,'The program does not carry out estimation for these data.');
   stopthis('Less than 2 relation types in this data.',
            outf, outf1, outfn, outfn1);
   exit;         
  end; (* if *)
 if (r > cmax) then begin
   writeln(outf,'The number of relation types is ',r,'.');
   writeln(outf,'This is more than constant cmax = ', cmax,' in B_CONS.');
   writeln(outf,'Recompile BLOCKS with larger constant cmax.');
   stopthis('Too many relation types in this data.',
            outf, outf1, outfn, outfn1);
   exit;         
  end; (* if *)

end; (* readdata *)

procedure Blockmodel.IdentifyVertices(var outf: textfile);
(* Identification of colors through priors for x[i].
   This procedure defines arrays prior and thetai. *)
var i,j: integer;
  ncol: array[1..cmax] of integer;
  col: array of integer;
(* Col[i] is the color for vertex i ;
   ncol[i] is the number of vertices for color i. *)
begin
  if IdentificationByFile then exit;
  (* Then the definition was given already. *)
  writeln(outf);
  Heading(outf,2,'Vertex identification for identified model.');
  SetLength(col,n+1);
  for i:=1 to n do begin
    col[i] := 0;
    prior[i] := false;
    for j:=1 to cmax do thetai[i,j] := 0.0;
   end;
  for i:=1 to cmax do ncol[i] := 0;

  if both then begin
    for i:=1 to c do begin
      col[colver[i]] := i;
      ncol[i] := 1;
     end; (* for i *)
    if (lastthreshold > 0.1) then begin
      col[colver[c]] := 0;
      ncol[c] := 0;
     end; (* if *)
  end (* both*) else begin
   for i:=1 to c do begin
     if ((v[i] >= 1) and (v[i] <= n)) then begin
       ncol[i] := 1;
       col[v[i]] := i;
      end;
    end; (* for i *)
  end; (* not both *)

  for i:=1 to n do begin
    if (col[i] > 0) then
      writeln(outf,'Vertex ',i,' is used to identify color ',col[i],'.');
   end;

  for i:=1 to n do if (col[i] > 0) then begin
    prior[i] := true;
    if (ncol[col[i]] = 1) then begin
      for j:=1 to c do thetai[i,j] := 0.05/(c-1);
      thetai[i,col[i]] := 0.95;
     end; (* if *)
    if (ncol[col[i]] >= 2) then begin
      for j:=1 to c do thetai[i,j] := 0.1/(c-1);
      thetai[i,col[i]] := 0.9;
    end; (* if *)
  end (* for i if (col[i] > 0) *) else prior[i] := false;

  writeln(outf);
  writeln(outf,'The following vertices have prior probabilities for the colors:');
  writeln(outf,'           colors');
  write(outf,  'vertex  ');
  for j:=1 to c do write(outf,j:6); writeln(outf);
  for i:=1 to n do if prior[i] then begin
    write(outf,i:3, '       ');
    for j:=1 to c do write(outf,thetai[i,j]:6:3); writeln(outf);
   end; (* i *)
  writeln(outf);
  finalize(col);
end; (* IdentifyVertices *)

begin (* unit BlockMo *)
  version := '1.8';
(* Changes in 1.52 compared to 1.51: unimportant output changes. *)
(* Change in 1.53:
     possibility for range check error in StrictBlock corrected. *)
(* Change in 1.6: unlimited number of vertices
    by using dynamic arrays *)
(* Change in 1.7: Pajek output in StrictBloks *)
(* Change in 1.8: Longer overdispersed burn-in. *)
end.

